#include "stdio.h"
#include "stdlib.h"

#include "iRRAM_core.h"

/*****************************************************************/
/*                     sin(x), cos(x), tan(x)  etc               */
/*****************************************************************/


REAL sin_approx (long prec, long wd, const REAL& x){
  REAL x2=square(x);
  REAL xpow[wd],xf=power(x2,wd),z=x;
  xpow[0]=x;
  for (long i=1; i<wd; i++) {
     xpow[i]=xpow[i-1]*x2/(-2*i*(2*i+1));
     z+=xpow[i];
  }
  REAL xp=1,e=z;
  INTEGER f=1;
  long i=1;
  long t=upperbound(xf),t1=t;
  while ( !bound(e,prec) || ! bound(e,-1) ) {
    xp*=xf; e=0;
    for (long j=0; j<wd; j++) {
      f=1;
      for (long k=i;k<i+wd;k++) f*=(-2*k*(2*k+1));
      xpow[j]=scale(xpow[j],t)/REAL(f);
      i=i+1;
      e+=xpow[j];
    }
    e*=scale(xp,-t1); t1=t1+t;
    z+=e;
  }
//  {sizetype err; e.geterror(err);z.adderror(err);}
  return z;
};

REAL sin3 (long prec,const REAL& x){
  long it;
  long wd;
  if ( prec > -100 ) {it =4; wd= 1;}
  else if ( prec > -400 ) {it =6; wd= 1;}
  else if (prec > -800 ) {it=10; wd =2;}
  else if (prec > -1600 ) {it=12; wd =2;}
  else if (prec > -3200 ) {it=14; wd =3;}
  else if (prec > -6400 ) {it=16; wd =4;}
  else if (prec > -12000 ) {it=18; wd =5;}
  else if (prec > -24000 ) {it=20; wd =8;}
  else if (prec > -48000 ) {it=24; wd =12;}
  else if (prec > -100000 ) {it=26; wd =18;}
  else if (prec > -200000 ) {it=28; wd =26;}
  else {
     it=(long)(log(double(-prec))*2+3);
     wd=long(sqrt(double(-prec)/18));
  }  
  REAL y=x;
  long jt=it;
  while (jt>=10) {y=y/59049;jt=jt-10;};
  while (jt>=3)  {y=y/27;jt=jt-3;};
  while (jt>=1)  {y=y/3;jt=jt-1;};
  y=sin_approx(prec,wd,y);
  for (long i=1;i<=it;i+=1) y*=(3-scale(square(y),2));
  return y;
};


REAL cos1 (long prec,const REAL& x){
  if (prec >= 0 ) return 0;
  REAL m=modulo(x,2*pi());
  long i=round(m*100);
  if ( i <  0 ) {i=i+628;m=m+2*pi();}
  if ( i> 550 )
    m= sqrt(1-square(sin3(prec,m-2*pi())));
  else if ( i> 392)
    m= sin3(prec,m-3*pi()/2);
  else if ( i> 235)
    m= -sqrt(1-square(sin3(prec,pi()-m)));
  else if ( i> 80)
    m= sin3(prec,pi()/2-m);
  else
    m= sqrt(1-square(sin3(prec,m)));
  return m;
};

REAL sin1 (long prec,const REAL& x){
  if (prec >= 0 ) return 0;
  REAL m=modulo(x,2*pi());
  long i=round(m*100);
  if ( i <  0 ) {i=i+628;m=m+2*pi();}
  if ( i> 550 )
    m= sin3(prec,m-2*pi());
  else if ( i> 392)
    m= -sqrt(1-square(sin3(prec,m-3*pi()/2)));
  else if ( i> 235)
    m= sin3(prec,pi()-m);
  else if ( i> 80)
    m= sqrt(1-square(sin3(prec,pi()/2-m)));
  else
    m= sin3(prec,m);
  return m;
};

REAL sin (const REAL& x){ return  limit_lip(sin1,0,x);};

REAL cos (const REAL& x){ return  limit_lip(cos1,0,x);};


REAL tan (const REAL& x){
   return sin(x)/cos(x);
};

///////////////////////////////////////////////////////////////////

REAL atan_approx (long prec, const REAL& x){
   if ( prec >= 2 ) return 0;
   long i=2;
   REAL x2=x*x,y=-x*x2,z=x+y/3;
   while ( !bound(y,prec) || ! bound(y,-1) ) {
     y=-y*x2;
     z=z+y/(i+3);
     i+=2;
   }
   return z;
};

REAL atan (const REAL& x){
  if ( positive(0.01-abs(x),-10) )
    return limit_lip(atan_approx,2,x);
  if ( positive(2-abs(x),-10) ) { 
    REAL y= (sqrt(x*x+1)-1)/x;
    y=  (sqrt(y*y+1)-1)/y;
    return 8*limit_lip(atan_approx,2,(sqrt(y*y+1)-1)/y);
    }
  if ( x > 0   ) return   pi()/2 - atan(1/x);
  else           return -(pi()/2 - atan(-1/x));

};

REAL asin (const REAL& x){
   return atan(x/sqrt(1-x*x));
};

REAL acos (const REAL& x){
   return pi()/2 - asin(x);
};

REAL asec (const REAL& x){
   return acos(1/x);
};

REAL acosec (const REAL& x){
   return asin(1/x);
};

///////////////////////////////////////////////////////////////////
REAL sec (const REAL& x) {
 return 1/cos(x);
}

REAL cosec (const REAL& x) {
 return 1/sin(x);
}

REAL cotan (const REAL& x) {
 return cos(x)/sin(x);
}

REAL sech (const REAL& x) {
 return 2/(exp(x)+exp(-x));
}

REAL cosech (const REAL& x) {
 return 2/(exp(x)-exp(-x));
}

REAL sinh (const REAL& x) {
 return (exp(x)-exp(-x))/2;
}

REAL cosh (const REAL& x) {
 return (exp(x)+exp(-x))/2;
}

REAL tanh (const REAL& x) {
 return (exp(x)-exp(-x))/(exp(x)+exp(-x));
}

REAL coth (const REAL& x) {
 return (exp(x)+exp(-x))/(exp(x)-exp(-x));
}

REAL asinh (const REAL& x) {
 return log(x+sqrt(x*x+1));
}

REAL acosh (const REAL& x) {
 return log(x+sqrt(x*x-1));
}

REAL atanh (const REAL& x) {
 return log((1+x)/(1-x))/2;
}

REAL acoth (const REAL& x) {
 return log((1-x)/(1+x))/2;
}


REAL asech (const REAL& x){
   return acosh(1/x);
};

REAL acosech (const REAL& x){
   return asinh(1/x);
};



