/*

stack.cc -- basic file for the iterations in the iRRAM library
 
Copyright (C) 2001 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

#include <math.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "iRRAM.h"
#include "iRRAM_core.h"
#include "iRRAM_version.h"

//*************************************************************************************
// constructor ITERATION_STACK
//*************************************************************************************


ITERATION_STACK::ITERATION_STACK() {
prec_diff=0;
prec_policy=0;
//printf("Neuer Stack angelegt\n");
}

//*************************************************************************************
// destructor ITERATION_STACK
//*************************************************************************************

ITERATION_STACK::~ITERATION_STACK()
{
//fprintf(stderr,"Destruktor ITERATION_STACK\n");
}

ITERATION_STACK * ACTUAL_STACK;

/* Implementation of class METRIC_OBJECT ********************/

ITERATION_STACK MAIN_STACK;
long int METRIC_OBJECT_COUNT=0;

METRIC_OBJECT::METRIC_OBJECT()
{
METRIC_OBJECT_COUNT+=1;      
//fprintf(stderr," +- Erzeuge: %x\n",this);
}

METRIC_OBJECT::~METRIC_OBJECT()
{
//fprintf(stderr," +- Lsche: %x\n",this);
  METRIC_OBJECT_COUNT-=1; 
}

/*****************************************/
// iRRAM_exec
 

void iRRAM_exec(void iRRAM()){
position=eingabe;
eingabe[0]=0;
printfausgaben=0;
printfaufrufe=0;
testaufrufe=0;
testanzahl=0;
konvertanzahl=0;ikonvertanzahl=0;rkonvertanzahl=0;
konvertaufrufe=0;ikonvertaufrufe=0;rkonvertaufrufe=0;

prec_inc=prec_inc0;
actual_prec=starting_prec+prec_inc; 

ACTUAL_STACK = &MAIN_STACK;

DEBUG3(1,"\niRRAM (version %s, backend %s) starting...\n",iRRAM_VERSION_rt,iRRAM_BACKEND);

inlimit=0;

while (true) {
  long p_end;
  try { iRRAM();   if (!infinite) break;}
  catch ( Iteration it)  { p_end=actual_prec+it.prec_diff; }

    	do {
      		prec_inc=long(prec_inc * prec_factor)+prec_inc1;
      		actual_prec=starting_prec+prec_inc;
    	} while ( actual_prec > p_end ); 
    	inlimit=0;
    	DEBUG2(1,"   objects in use:     %ld\n",METRIC_OBJECT_COUNT); 
	DEBUG2(1,"   MP-objects in use:  %ld\n",MP_var_count); 
#ifdef	MP_space_count
    	DEBUG2(1,"   MP-memory in use:   %ld kB\n",MP_space_count); 
    	DEBUG2(1,"   max MP-memory used: %ld kB\n",MP_max_space_count); 
#endif
    	DEBUG2(1,"   cached tests:       %ld \n",testanzahl); 
	DEBUG2(1,"   cached conversions: %ld \n",konvertanzahl); 
	DEBUG2(1,"   cached inputs:      %d \n",position-eingabe); 
	DEBUG2(1,"   number of outputs:  %ld \n",printfaufrufe);
    	position=eingabe;
	printfaufrufe=0;
	testaufrufe=0;
	DEBUG2(1,"increasing precision bound to 2^(%ld)...\n",actual_prec);
  }


DEBUG1(1,"iRRAM ending \n");
DEBUG2(1,"   objects in use:     %ld\n",METRIC_OBJECT_COUNT); 
DEBUG2(1,"   MP-objects in use:  %ld\n",MP_var_count); 
#ifdef	MP_space_count
DEBUG2(1,"   MP-memory in use:   %ld kB\n",MP_space_count); 
DEBUG2(1,"   max MP-memory used: %ld kB\n",MP_max_space_count); 
#endif
DEBUG2(1,"   cached tests:       %ld \n",testanzahl); 
DEBUG2(1,"   cached conversions: %ld \n",konvertanzahl); 
DEBUG2(1,"   cached inputs:      %d \n",position-eingabe); 
DEBUG2(1,"   number of outputs:  %ld \n",printfaufrufe); 
DEBUG2(1,"   maximal precision:  %ld \n",actual_prec); 

for (long i= 0; i < konvertanzahl; i++) {
  MP_clear(konvert[i]);
  konvert[i]=0;
  }

//************************** neu *******************************************

for (long i= 0; i < ikonvertanzahl; i++) {
  MP_int_clear(ikonvert[i]);
  ikonvert[i]=0;
  }
for (long i= 0; i < rkonvertanzahl; i++) {
  MP_rat_clear(rkonvert[i]);
  rkonvert[i]=0;
  }

rkonvertanzahl=0;
ikonvertanzahl=0;
konvertanzahl=0;

DEBUG1(1,"\n   after cleaning up:\n");
DEBUG2(1,"   objects in use:     %ld\n",METRIC_OBJECT_COUNT); 
DEBUG2(1,"   MP-objects in use:  %ld\n",MP_var_count); 

}


/******************************EOF********************************************/

